from __future__ import annotations


class SqlRenderingConstants:
    """Constants used for rendering SQL."""

    # The indentation for SQL e.g.
    # SELECT
    #   foo
    #   , bar
    INDENT = "  "
